/* libnormaliz/nmz_config.h.in.  Maintained manually.
   libnormaliz/nmz_config.h generated by configure. */

/* Defines the preprocessor variables that control optional packages */
#undef ENFNORMALIZ
#undef NMZ_HASHLIBRARY
#undef NMZ_NAUTY
#undef NMZ_NAUTYNAUTY
#undef NMZ_NAUTY_TLS
#undef NMZ_FLINT
#undef NMZ_COCOA

#define ENFNORMALIZ

#define NMZ_NAUTY
#define NMZ_NAUTYNAUTY

#define NMZ_FLINT

#define NORMALIZ_USE_DLL
